import logging

from .blocks import ChildDatabaseBlock, ChildPageBlock

from n2y.utils import fromisoformat
from n2y.property_values import TitlePropertyValue


logger = logging.getLogger(__name__)


class Page:
    def __init__(self, client, notion_data):
        logger.debug("Instantiating page")
        self.client = client

        self.notion_data = notion_data
        self.notion_id = notion_data['id']
        self.created_time = fromisoformat(notion_data['created_time'])
        self.created_by = client.wrap_notion_user(notion_data['created_by'])
        self.last_edited_time = fromisoformat(notion_data['last_edited_time'])
        self.last_edited_by = client.wrap_notion_user(notion_data['last_edited_by'])
        self.archived = notion_data['archived']
        self.emoji = self._init_icon(notion_data['icon'])
        self.cover = notion_data['cover'] and client.wrap_notion_file(notion_data['cover'])
        self.archived = notion_data['archived']
        self.properties = {
            k: client.wrap_notion_property_value(npv)
            for k, npv in notion_data['properties'].items()
        }
        self.notion_parent = notion_data['parent']
        self.notion_url = notion_data['url']

        self._block = None
        self._children = None

        self.plugin_data = {}

    def _init_icon(self, icon_notion_data):
        """
        The icon property is unique in that it can be either an emoji or a file.
        """
        if icon_notion_data is None:
            return None
        elif icon_notion_data["type"] == "emoji":
            return self.client.wrap_notion_emoji(icon_notion_data)
        else:
            return self.client.wrap_notion_file(icon_notion_data)

    @property
    def title(self):
        for property_value in self.properties.values():
            # Notion ensures there is always exactly one title property
            if isinstance(property_value, TitlePropertyValue):
                return property_value.rich_text

    @property
    def block(self):
        if self._block is None:
            self._block = self.client.get_block(self.notion_id, page=self)
        return self._block

    @property
    def children(self):
        """
        Get a list of child pages and databases.
        """
        if self._children is None:
            self._children = []
            for block in self.block.children:
                self._append_children(block)
        return self._children

    def get_children(self):
        self.block.get_children()
        self._children = []
        for block in self.block.children:
            self._append_children(block)

    def _append_children(self, block):
        if isinstance(block, ChildPageBlock):
            page = self.client.get_page(block.notion_id)
            self._children.append(page)
        elif isinstance(block, ChildDatabaseBlock):
            database = self.client.get_database(block.notion_id)
            self._children.append(database)
        elif block.children is not None:
            # Recursively look for child pages and databases in the hierarchy
            for child_block in block.children:
                self._append_children(child_block)

    @property
    def parent(self):
        parent_type = self.notion_parent["type"]
        if parent_type == "workspace":
            return None
        elif parent_type == "page_id":
            return self.client.get_page(self.notion_parent["page_id"])
        else:
            assert parent_type == "database_id"
            return self.client.get_database(self.notion_parent["database_id"])

    def to_pandoc(self):
        return self.block.to_pandoc()

    def properties_to_values(self):
        return {k: v.to_value() for k, v in self.properties.items()}



"""
This module contains all the code responsible for exporting `page.Page` and
`database.Database` objects into the various supported file formats.
"""
import os
import logging

import yaml

from n2y.utils import pandoc_write_or_log_errors, sanitize_filename

logger = logging.getLogger(__name__)


def _page_properties(page, id_property=None, url_property=None, property_map=None):
    if property_map is None:
        property_map = {}
    properties = page.properties_to_values()
    if id_property in properties:
        logger.warning(
            'The id property "%s" is shadowing an existing '
            'property with the same name', id_property,
        )
    if id_property:
        properties[id_property] = page.notion_id

    if url_property in properties:
        logger.warning(
            'The url property "%s" is shadowing an existing '
            'property with the same name', url_property,
        )
    if url_property:
        properties[url_property] = page.notion_url
    for original, new in property_map.items():
        if original in properties:
            properties[new] = properties.pop(original)
        else:
            msg = "Property %s not found in page %s; skipping remapping from %s to %s"
            logger.warning(msg, original, page.notion_url, original, new)
    return properties


def export_page(
    page,
    pandoc_format,
    pandoc_options,
    id_property=None,
    url_property=None,
    property_map=None,
):
    page_properties = _page_properties(page, id_property, url_property, property_map)
    pandoc_ast = page.to_pandoc()
    page_content = pandoc_write_or_log_errors(pandoc_ast, pandoc_format, pandoc_options)
    return '\n'.join([
        '---',
        yaml.dump(page_properties) + '---',
        page_content,
    ])


def database_to_yaml(
    database,
    pandoc_format,
    pandoc_options,
    notion_filter=None,
    notion_sorts=None,
    id_property=None,
    url_property=None,
    content_property=None,
    property_map=None,
):
    if content_property in database.schema:
        logger.warning(
            'The content property "%s" is shadowing an existing '
            'property with the same name', content_property,
        )
    results = []
    for page in database.children_filtered(notion_filter, notion_sorts):
        result = _page_properties(page, id_property, url_property, property_map)
        if content_property:
            pandoc_ast = page.to_pandoc()
            if pandoc_ast:
                result[content_property] = pandoc_write_or_log_errors(
                    pandoc_ast, pandoc_format, pandoc_options,
                )
            else:
                result[content_property] = None
        results.append(result)
    return yaml.dump(results, sort_keys=False)


def database_to_markdown_files(
    database,
    directory,
    pandoc_format,
    pandoc_options,
    filename_property=None,
    notion_filter=None,
    notion_sorts=None,
    id_property=None,
    url_property=None,
    property_map=None,
):
    os.makedirs(directory, exist_ok=True)
    seen_file_names = set()
    counts = {'unnamed': 0, 'duplicate': 0}
    for page in database.children_filtered(notion_filter, notion_sorts):
        page_filename = _page_filename(page, filename_property)
        if page_filename:
            if page_filename not in seen_file_names:
                seen_file_names.add(page_filename)
                with open(os.path.join(directory, f"{page_filename}.md"), 'w') as f:
                    document = export_page(
                        page,
                        pandoc_format,
                        pandoc_options,
                        id_property,
                        url_property,
                        property_map,
                    )
                    f.write(document)
            else:
                logger.warning('Skipping page named "%s" since it has been used', page_filename)
                counts['duplicate'] += 1
        else:
            counts['unnamed'] += 1
    for key, count in counts.items():
        if count > 0:
            logger.info("%d %s page(s) skipped", count, key)


def _page_filename(page, filename_property):
    # TODO: switch to using the database's natural keys as the file names
    if filename_property is None:
        return sanitize_filename(page.title.to_plain_text())
    elif filename_property in page.properties:
        return sanitize_filename(page.properties[filename_property].to_value())
    else:
        logger.warning(
            'Invalid filename property, "%s". Valid options are %s',
            filename_property, ", ".join(page.properties.keys()),
        )
        return sanitize_filename(page.title.to_plain_text())



import os
import sys
import logging
import argparse
import pkg_resources

from n2y import notion
from n2y.export import export_page, database_to_yaml, database_to_markdown_files
from n2y.config import load_config
from n2y.utils import share_link_from_id

logger = None


def cli_main():
    args = sys.argv[1:]
    access_token = os.environ.get('NOTION_ACCESS_TOKEN', None)
    sys.exit(main(args, access_token))


def main(raw_args, access_token):
    parser = argparse.ArgumentParser(
        description="Move data from Notion into YAML/markdown",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument("config", help="The path to the config file")
    parser.add_argument(
        "--verbosity", '-v', default='INFO',
        help="Level to set the root logging module to",
    )
    parser.add_argument(
        "--version", action='version', version=pkg_resources.require("n2y")[0].version,
        help="The version of n2y installed",
    )

    args = parser.parse_args(raw_args)

    logging_level = logging.__dict__[args.verbosity]
    logging.basicConfig(level=logging_level)
    global logger
    logger = logging.getLogger(__name__)

    if access_token is None:
        logger.critical('No NOTION_ACCESS_TOKEN environment variable is set')
        return 1

    config = load_config(args.config)
    if config is None:
        return 2

    client = notion.Client(access_token, config["media_root"], config["media_url"])

    error_occurred = False
    for export in config['exports']:
        logger.info("Exporting to %s", export['output'])
        client.load_plugins(export["plugins"])
        export_completed = _export_node_from_config(client, export)
        if not export_completed:
            error_occurred = True
    return 0 if not error_occurred else 3


def _export_node_from_config(client, export):
    node_type = export["node_type"]
    if node_type == "page":
        page = client.get_page(export['id'])
        if page is None:
            msg = (
                "Unable to find page with id '%s' (%s). "
                "Perhaps the integration doesn't have permission to access this page?"
            )
            logger.error(msg, export['id'], share_link_from_id(export['id']))
            return False
        result = export_page(
            page,
            export["pandoc_format"],
            export["pandoc_options"],
            export["id_property"],
            export["url_property"],
            export["property_map"],
        )
        with open(export["output"], "w") as f:
            f.write(result)
    else:
        database = client.get_database(export['id'])
        if database is None:
            msg = (
                "Unable to find database with id '%s' (%s). "
                "Perhaps the integration doesn't have permission to access this database?"
            )
            logger.error(msg, export['id'], share_link_from_id(export['id']))
            return False
        if node_type == "database_as_yaml":
            result = database_to_yaml(
                database=database,
                pandoc_format=export["pandoc_format"],
                pandoc_options=export["pandoc_options"],
                id_property=export["id_property"],
                url_property=export["url_property"],
                content_property=export["content_property"],
                notion_filter=export["notion_filter"],
                notion_sorts=export["notion_sorts"],
                property_map=export["property_map"],
            )
            with open(export["output"], "w") as f:
                f.write(result)
        elif node_type == "database_as_files":
            database_to_markdown_files(
                database=database,
                directory=export["output"],
                pandoc_format=export["pandoc_format"],
                pandoc_options=export["pandoc_options"],
                filename_property=export["filename_property"],
                notion_filter=export["notion_filter"],
                notion_sorts=export["notion_sorts"],
                id_property=export["id_property"],
                url_property=export["url_property"],
                property_map=export["property_map"],
            )
        else:
            logger.error("Unknown node_type '%s'", node_type)
            return False
    return True


if __name__ == "__main__":
    cli_main()
